#include "bayesian/bayesianestimator.h"
#include "evaluation/error.h"
#include "common/color.h"
#include "common/mask.h"
#include "common/scale.h"
#include "common/statistics.h"
#include <opencv2/highgui/highgui.hpp>
#include <opencv2/imgproc/imgproc.hpp>
#include <math.h>
#include <limits>
#include <iostream>
#include <vector>
#include <set>

namespace illumestimators {

BayesianEstimator::BayesianEstimator(Method method, int verbosity, bool cache) :
	m_method(method),
	m_size(32),
	verbosity(verbosity),
	m_cache(cache)
{
	const int sizes[] = {m_size, m_size, m_size};
	m_logReflectancePrior.create(3, sizes);
	m_logReflectancePrior = 0;
}

BayesianEstimator::BayesianEstimator(Method  method, const std::vector<cv::Vec3d>& illuminants, const cv::Mat_<double>& logReflectancePrior, int verbosity, bool cache) :
	m_method(method),
	m_illuminants(illuminants),
	m_logReflectancePrior(logReflectancePrior.clone()),
	m_size(32),
	verbosity(verbosity),
	m_cache(cache)
{
	const int sizes[] = {m_size, m_size, m_size};
	m_logReflectancePrior.create(3, sizes);
	m_logReflectancePrior = 0;
}

BayesianEstimator::~BayesianEstimator()
{

}

std::string BayesianEstimator::name() const
{
	std::stringstream name;
	name << "BayesianEstimator(method = " << m_method << ")";

	return name.str();
}

std::string BayesianEstimator::identifier() const
{
	return std::string("bayesian");
}

double BayesianEstimator::sumlog(const std::vector<double>& values) const
{
	const double max = Statistics::max(values);
	double sum = 0;

	for (unsigned int i = 0; i < values.size(); i++) {
		// PATCH TO AVOID NAN  // PATCH TO AVOID NAN
		if ((values[i] == values[i]) && (fabs(values[i]) != std::numeric_limits<double>::infinity()))
			sum += exp(values[i] - max);
	}

	return max + log(sum);
}

cv::Mat_<int> BayesianEstimator::calculateHistogram(const std::vector<cv::Vec3d>& pixels) const
{
	const int sizes[] = {m_size, m_size, m_size};
	cv::Mat_<int> histogram(3, sizes, 0);

	for (unsigned int i = 0; i < pixels.size(); i++) {
		int index[3];

		for (int c = 0; c < 3; c++) {
			index[c] = histogram.size[c] * pixels[i][c];

			if (index[c] >= histogram.size[c]) {
				index[c] = histogram.size[c] - 1;
			} else if (index[c] < 0) {
				index[c] = 0;
			}
		}

		histogram(index[0], index[1], index[2])++;
	}

	return histogram;
}

std::vector<cv::Vec3d> BayesianEstimator::calculateReflectances(const std::vector<cv::Vec3d>& pixels, const cv::Vec3d& illuminant) const
{
	std::vector<cv::Vec3d> reflectances;
	reflectances.resize(pixels.size());

	if (fabs(illuminant[0] * illuminant[1] * illuminant[2]) <= std::numeric_limits<double>::epsilon()) return reflectances;
	for (unsigned int i = 0; i < pixels.size(); i++) {
		for (int c = 0; c < 3; c++) {
			reflectances[i][c] = pixels[i][c] / illuminant[c];
		}
	}

	return reflectances;
}

double BayesianEstimator::logLikelihoodRosenberg(const std::vector<cv::Vec3d>& pixels, const cv::Vec3d& illuminant) const
{
	std::vector<cv::Vec3d> reflectances = calculateReflectances(pixels, illuminant);

	cv::Mat_<int> histogram = calculateHistogram(reflectances);

	if (histogram.dims != m_logReflectancePrior.dims) {
		std::cerr << "Bayesian estimator: Histogram dimensions mismatch!" << std::endl;
		return 0;
	}

	for (int i = 0; i < histogram.dims; i++) {
		if (histogram.size[i] != m_logReflectancePrior.size[i]) {
			std::cerr << "Bayesian estimator: Histogram dimensions mismatch!" << std::endl;
			return 0;
		}
	}

	double sum = 0;
	int counter = 0;

	for (int i = 0; i < histogram.size[0]; i++) {
		for (int j = 0; j < histogram.size[1]; j++) {
			for (int k = 0; k < histogram.size[2]; k++) {
				if (histogram(i, j, k) > 0) {
					double bla = m_logReflectancePrior(i, j, k);
					if ((bla == bla) && (fabs(bla) != std::numeric_limits<double>::infinity())) {
						sum += bla;
						counter++;
					}
				}
			}
		}
	}

	const double size = reflectances.size();
	const double logPrior = size * (sum / (double) counter);
	const double logLikelihood = (-size * log(illuminant[0] * illuminant[1] * illuminant[2])) + logPrior;

	return logLikelihood;
}

void BayesianEstimator::preprocessImage(const cv::Mat_<cv::Vec3d>& image, const cv::Mat_<unsigned char>& mask, cv::Mat_<cv::Vec3d>& inputImage, cv::Mat_<unsigned char>& inputMask) const
{
	inputImage = image.clone();
	if ((image.rows != mask.rows) || (image.cols != mask.cols)) {
		std::cerr << "No mask!" << std::endl;
		inputMask = cv::Mat_<unsigned char>::zeros(inputImage.rows, inputImage.cols);
	} else {
		inputMask = mask.clone();
	}

	Mask::maskSaturatedPixels(inputImage, inputMask, 1);
	Mask::maskDarkPixels(inputImage, inputMask, 8.0 / 255);

//	const int scale = 3;
	const int scale = 1;

	cv::Mat_<cv::Vec3d> tmpImage;
	cv::blur(inputImage, tmpImage, cv::Size(scale, scale));

	cv::Mat_<unsigned char> tmpMask = inputMask;

	inputImage = cv::Mat_<cv::Vec3d>(tmpImage.rows / scale, tmpImage.cols / scale);
	inputMask = cv::Mat_<unsigned char>(tmpImage.rows / scale, tmpImage.cols / scale);

	for (int y = 0; y < inputImage.rows; y++) {
		for (int x = 0; x < inputImage.cols; x++) {
			inputImage(y, x) = tmpImage(y * scale, x * scale);
			inputMask(y, x) = tmpMask(y * scale, x * scale);
		}
	}
}

void BayesianEstimator::estimateIlluminants(
	const cv::Mat_<cv::Vec3d>& image,
	const std::vector<superpixels::Superpixel>& superpixels,
	std::vector<Illum> &estimated_illuminants,
	const cv::Mat_<unsigned char> &mask
	)
{
	estimated_illuminants = std::vector<Illum>(superpixels.size(), Illum());

	for (unsigned int s = 0; s < superpixels.size(); ++s) {
		if (verbosity > 0) std::cout << "estimating on superpixel " << s << std::endl;

		std::vector<cv::Vec3d> illuminants;
		std::vector<double> likelihoods;
		Illum estimate;

		const std::vector<cv::Point> &coord = superpixels[s].coordinates;
		cv::vector<cv::Vec3d> pixels(coord.size());
		int count = 0;
		for (int i = 0; i < (int)coord.size(); ++i) {
			if (mask[coord[i].y][coord[i].x] > 0) continue;
			pixels[i] = image[ coord[i].y ][ coord[i].x ];
			count++;
		}
		pixels.resize(count);
		const cv::Vec3d maxBGR = Statistics::max(pixels);

		if (verbosity > 0) std::cout << "m_illuminants.size() = " << m_illuminants.size() << std::endl;

		for (unsigned int i = 0; i < m_illuminants.size(); i++) {
			for (double brightness = 0.1; brightness < 4; brightness += 0.1) {
				cv::Vec3d illuminant;
				illuminant[2] = m_illuminants[i][2] * brightness;
				illuminant[1] = m_illuminants[i][1] * brightness;
				illuminant[0] = brightness - illuminant[1] - illuminant[2];

				if ((illuminant[0] < maxBGR[0]) || (illuminant[1] < maxBGR[1]) || (illuminant[2] < maxBGR[2])) {
					continue;
				}

				const double likelihood = logLikelihoodRosenberg(pixels, illuminant);

				illuminants.push_back(illuminant);
				likelihoods.push_back(likelihood);
			}
		}

		estimate = perform_actual_estimation(illuminants, likelihoods);
		estimated_illuminants[s] = Color::calculateChromaticities(estimate);
	}
}

Illum BayesianEstimator::estimateIlluminant(const cv::Mat_<cv::Vec3d>& image, const cv::Mat_<unsigned char>& mask) const
{
	if ((image.rows != mask.rows) || (image.cols != mask.cols)) {
		std::cerr << "Bayesian estimator: Mask size does not match image size!" << std::endl;
		return Illum();
	}

	if ((m_illuminants.size() < 1) || (m_logReflectancePrior.dims != 3)) {
		std::cerr << "Bayesian estimator: Training data missing or in wrong format!" << std::endl;
		return Illum();
	}

	cv::Mat_<cv::Vec3d> inputImage;
	cv::Mat_<unsigned char> inputMask;
	preprocessImage(image, mask, inputImage, inputMask);

	std::vector<cv::Vec3d> pixels = Mask::unmaskedPixels(inputImage, inputMask);
	const cv::Vec3d maxBGR = Statistics::max(pixels);

	std::vector<cv::Vec3d> illuminants;
	std::vector<double> likelihoods;

	if (verbosity > 0) std::cout << "iterating over " << m_illuminants.size() << " illuminants, 30 brightness each" << std::endl;
	for (unsigned int i = 0; i < m_illuminants.size(); i++) {
		if (verbosity > 0) if ((i % 2)  == 0) std::cout << "i = " << i << std::endl;
		for (double brightness = 0.1; brightness < 6; brightness += 0.1) {
			cv::Vec3d illuminant;
			illuminant[2] = m_illuminants[i][2] * brightness;
			illuminant[1] = m_illuminants[i][1] * brightness;
			illuminant[0] = brightness - illuminant[1] - illuminant[2];

			if ((illuminant[0] < maxBGR[0]) || (illuminant[1] < maxBGR[1]) || (illuminant[2] < maxBGR[2])) {
				continue;
			}

			const double likelihood = logLikelihoodRosenberg(pixels, illuminant);

			illuminants.push_back(illuminant);
			likelihoods.push_back(likelihood);
		}
	}

	Illum estimate = perform_actual_estimation(illuminants, likelihoods);
	return Color::calculateChromaticities(estimate);
}

Illum BayesianEstimator::estimateIlluminant(const cv::Mat_<cv::Vec3d>& image, const superpixels::Superpixel &superpixel, const cv::Mat_<unsigned char>& mask) const
{

	if ((image.rows != mask.rows) || (image.cols != mask.cols)) {
		std::cerr << "Bayesian estimator: Mask size does not match image size!" << std::endl;
		return Illum();
	}

	if ((m_illuminants.size() < 1) || (m_logReflectancePrior.dims != 3)) {
		std::cerr << "Bayesian estimator: Training data missing or in wrong format!" << std::endl;
		return Illum();
	}

	cv::Mat_<cv::Vec3d> inputImage;
	cv::Mat_<unsigned char> inputMask;
	preprocessImage(image, mask, inputImage, inputMask);

	std::vector<cv::Vec3d> pixels = Mask::unmaskedPixels(inputImage, superpixel, inputMask);
	const cv::Vec3d maxBGR = Statistics::max(pixels);

	std::vector<cv::Vec3d> illuminants;
	std::vector<double> likelihoods;

	for (unsigned int i = 0; i < m_illuminants.size(); i++) {
		for (double brightness = 0.1; brightness < 6; brightness += 0.1) {
			cv::Vec3d illuminant;
			illuminant[2] = m_illuminants[i][2] * brightness;
			illuminant[1] = m_illuminants[i][1] * brightness;
			illuminant[0] = brightness - illuminant[1] - illuminant[2];

			if ((illuminant[0] < maxBGR[0]) || (illuminant[1] < maxBGR[1]) || (illuminant[2] < maxBGR[2])) {
				continue;
			}

			const double likelihood = logLikelihoodRosenberg(pixels, illuminant);

			illuminants.push_back(illuminant);
			likelihoods.push_back(likelihood);
		}
	}

	Illum estimate = perform_actual_estimation(illuminants, likelihoods);
	return Color::calculateChromaticities(estimate);
}

cv::Mat_<int> BayesianEstimator::calculateReflectancesHistogram(
	const cv::Mat_<cv::Vec3d> &img,
	const superpixels::Superpixel &superpixel,
	const cv::Vec3d& illuminant,
	const cv::Mat_<unsigned char> &mask
	)
{
	//preprocessImage(img, inputMask); // FIXME: do preprocessing earlier, because it should not be repeated for every superpixel
	const std::vector<cv::Point> &coord = superpixel.coordinates;
	cv::vector<cv::Vec3d> pixels(coord.size());
	int count = 0;
	for (unsigned int i = 0; i < coord.size(); ++i) {
		if (mask[coord[i].y][coord[i].x] > 0) continue;
		pixels[i] = img[ coord[i].y ][ coord[i].x ];
		count++;
	}
	pixels.resize(count);
	std::vector<cv::Vec3d> reflectances = calculateReflectances(pixels, illuminant);
	Scale::scaleByMax(reflectances);

	return calculateHistogram(reflectances);
}

cv::Mat_<int> BayesianEstimator::calculateReflectancesHistogram(const std::string& imageFile, const std::string& colorspace, const cv::Vec3d& illuminant, const std::string& maskFile)
{
	cv::Mat_<cv::Vec3d> image = cleanlyReadImage(imageFile, colorspace);
	cv::Mat_<unsigned char> mask = cv::imread(maskFile, 0);
	
	cv::Mat_<cv::Vec3d> inputImage;
	cv::Mat_<unsigned char> inputMask;
	preprocessImage(image, mask, inputImage, inputMask);

	std::vector<cv::Vec3d> pixels = Mask::unmaskedPixels(inputImage, inputMask);

	std::vector<cv::Vec3d> reflectances = calculateReflectances(pixels, illuminant);

	Scale::scaleByMax(reflectances);

	return calculateHistogram(reflectances);
}

bool BayesianEstimator::finalizeTraining()
{
	if (m_illuminants.size() == 0) {
		m_illuminants.insert(m_illuminants.end(), distinct_illums.begin(), distinct_illums.end());
	}

	double sum = 0;
	for (int j = 0; j < m_logReflectancePrior.size[0]; j++) {
		for (int k = 0; k < m_logReflectancePrior.size[1]; k++) {
			for (int l = 0; l < m_logReflectancePrior.size[2]; l++) {
				sum += m_logReflectancePrior(j, k, l);
			}
		}
	}

	if (fabs(sum) <= std::numeric_limits<double>::epsilon()) {
		if (verbosity > 0) std::cout << "ARGHS __ SUM IST NULL!" << std::endl;
	}

	for (int j = 0; j < m_logReflectancePrior.size[0]; j++) {
		for (int k = 0; k < m_logReflectancePrior.size[1]; k++) {
			for (int l = 0; l < m_logReflectancePrior.size[2]; l++) {
				if (fabs(m_logReflectancePrior(j, k, l)) <= std::numeric_limits<double>::epsilon()) {
					if (verbosity > 0) std::cout << "FAIL LOG OF ZERO!" << std::endl;
					m_logReflectancePrior(j, k, l) = -10000000 / sum;
				}  else {
					m_logReflectancePrior(j, k, l) = log(m_logReflectancePrior(j, k, l) / sum);
				}
			}
		}
	}

	// dump reflectance prior
	std::ofstream out;
	out.open("/tmp/reflectance_prior.txt");
	for (int j = 0; j < m_logReflectancePrior.size[0]; j++) {
		for (int k = 0; k < m_logReflectancePrior.size[1]; k++) {
			out << "j=" << j << ",k=" << k << ": ";
			for (int l = 0; l < m_logReflectancePrior.size[2]; l++) {
				out << m_logReflectancePrior(j, k, l) << " ";
			}
			out << std::endl;
		}
	}
	out.close();

	return true;
}

bool BayesianEstimator::addTrainingData(cv::Mat_<cv::Vec3d> img, const std::vector<superpixels::Superpixel>& superpixels, const std::vector<cv::Vec3d>& illuminants, cv::Mat_<unsigned char> &mask)
{
	if (superpixels.size() != illuminants.size()) {
		std::cerr << "Bayesian estimator: Training data dimensions mismatch!" << std::endl;
		return false;
	}

	if (superpixels.size() < 1) {
		std::cerr << "Bayesian estimator: Estimator needs at least one image for training!" << std::endl;
		return false;
	}

	distinct_illums.insert(illuminants.begin(), illuminants.end());
//	m_illuminants.insert(m_illuminants.end(), illuminants.begin(), illuminants.end());


	for (unsigned int i = 0; i < superpixels.size(); i++) {

		cv::Mat_<int> histogram = calculateReflectancesHistogram(img, superpixels[i], illuminants[i], mask);

		for (int j = 0; j < histogram.size[0]; j++) {
			for (int k = 0; k < histogram.size[1]; k++) {
				for (int l = 0; l < histogram.size[2]; l++) {
					if (histogram(j, k, l) > 0) {
						m_logReflectancePrior(j, k, l)++;
					}
				}
			}
		}
	}

	return true;
}





bool BayesianEstimator::train(const std::vector<std::string>& imageFiles, const std::vector<std::string>& colorspaces, const std::vector<cv::Vec3d>& illuminants, const std::vector<std::string>& maskFiles)
{
	if ((imageFiles.size() != colorspaces.size()) || (imageFiles.size() != illuminants.size())) {
		std::cerr << "Bayesian estimator: Training data dimensions mismatch!" << std::endl;
		return false;
	}

	if (imageFiles.size() < 1) {
		std::cerr << "Bayesian estimator: Estimator needs at least one image for training!" << std::endl;
		return false;
	}

	if ((maskFiles.size() > 0) && (maskFiles.size() != imageFiles.size())) {
		std::cerr << "Bayesian estimator: Mask data dimensions mismatch!" << std::endl;
		return false;
	}

	m_illuminants = illuminants;

	/* Now in Constructor
	const int sizes[] = {m_size, m_size, m_size};
	m_logReflectancePrior.create(3, sizes);
	m_logReflectancePrior = 0;
	*/

	m_logReflectancePrior = 0;

	for (unsigned int i = 0; i < imageFiles.size(); i++) {
		std::string maskFile;

		if (maskFiles.size() > i) {
			maskFile = maskFiles[i];
		}

		cv::Mat_<int> histogram;

		if (m_cache) {
			static std::map<std::string, cv::Mat_<int> > histogramCache;

			// TODO: Resolve proper path, load/save
			if (histogramCache.find(imageFiles[i]) != histogramCache.end()) {
				histogram = histogramCache[imageFiles[i]];
			} else {
				histogram = calculateReflectancesHistogram(imageFiles[i], colorspaces[i], illuminants[i], maskFile);
				histogramCache[imageFiles[i]] = histogram;
			}
		} else {
			histogram = calculateReflectancesHistogram(imageFiles[i], colorspaces[i], illuminants[i], maskFile);
		}

		for (int j = 0; j < histogram.size[0]; j++) {
			for (int k = 0; k < histogram.size[1]; k++) {
				for (int l = 0; l < histogram.size[2]; l++) {
					if (histogram(j, k, l) > 0) {
						m_logReflectancePrior(j, k, l)++;
					}
				}
			}
		}
	}

	finalizeTraining();
	return true;
}

bool BayesianEstimator::train(const std::vector<cv::Mat_<int> > histograms, const std::vector<cv::Vec3d>& illuminants)
{
	if (histograms.size() != illuminants.size()) {
		std::cerr << "Bayesian estimator: Training data dimensions mismatch!" << std::endl;
		return false;
	}

	if (histograms.size() < 1) {
		std::cerr << "Bayesian estimator: Estimator needs at least one histogram for training!" << std::endl;
		return false;
	}

	m_illuminants = illuminants;

	/* Now in Constructor
	const int sizes[] = {m_size, m_size, m_size};
	m_logReflectancePrior.create(3, sizes);
	m_logReflectancePrior = 0;
	*/

	m_logReflectancePrior = 0;

	for (unsigned int i = 0; i < histograms.size(); i++) {
		cv::Mat_<int> histogram = histograms[i];

		for (int j = 0; j < histogram.size[0]; j++) {
			for (int k = 0; k < histogram.size[1]; k++) {
				for (int l = 0; l < histogram.size[2]; l++) {
					if (histogram(j, k, l) > 0) {
						m_logReflectancePrior(j, k, l)++;
					}
				}
			}
		}
	}
	finalizeTraining();


	return true;
}

Illum BayesianEstimator::perform_actual_estimation(const std::vector<cv::Vec3d> &illuminants, const std::vector<double> &likelihoods) const
{
	cv::Vec3d estimate(true);
	if (m_method == POSTERIOR_MEAN) {
		const double likelihoodSum = sumlog(likelihoods);

		for (int c = 0; c < 3; c++) {
			std::vector<double> tmp;
			tmp.resize(illuminants.size());

			for (unsigned int i = 0; i < illuminants.size(); i++) {
				tmp[i] = log(illuminants[i][c]) + likelihoods[i] - likelihoodSum;
			}

			estimate[c] = exp(sumlog(tmp));
		}
	} else if (m_method == MIN_AVERAGE_LOSS_EUCLIDEAN) {
		std::vector<double> averageLoss;
		averageLoss.resize(illuminants.size());

		for (unsigned int i = 0; i < illuminants.size(); i++) {
			cv::Vec3d illuminant = illuminants[i];

			std::vector<double> losses;

			for(unsigned int j = 0; j < illuminants.size(); j++) {
				const double loss = log(Error(illuminant, illuminants[j]).toRgError()) + likelihoods[j];
				losses.push_back(loss);
			}

			averageLoss[i] = (sumlog(losses));
		}

		double min = std::numeric_limits<double>::max();
		for (unsigned int i = 0; i < illuminants.size(); i++) {
			if (averageLoss[i] < min) {
				min = averageLoss[i];
				estimate = illuminants[i];
			}
		}
	} else if (m_method == MIN_AVERAGE_LOSS_ANGULAR) {
		std::vector<double> averageLoss;
		averageLoss.resize(illuminants.size());

		for (unsigned int i = 0; i < illuminants.size(); i++) {
			cv::Vec3d illuminant = illuminants[i];

			std::vector<double> losses;

			for(unsigned int j = 0; j < illuminants.size(); j++) {
				const double loss = log(Error(illuminant, illuminants[j]).toAngularErrorRad()) + likelihoods[j];
				losses.push_back(loss);
			}

			averageLoss[i] = (sumlog(losses));
		}

		double min = std::numeric_limits<double>::max();
		for (unsigned int i = 0; i < illuminants.size(); i++) {
			if (averageLoss[i] < min) {
				min = averageLoss[i];
				estimate = illuminants[i];
			}
		}
	} else {
		std::cerr << "Bayesian estimator: Unknown method selected!" << std::endl;
		return Illum();
	}

	return Illum(estimate[2], estimate[1], estimate[0]); // assumes bgr order
}

bool BayesianEstimator::save(const std::string& filename) const
{
	cv::FileStorage fs(filename, cv::FileStorage::WRITE);

	if (!fs.isOpened()) {
		return false;
	}

	fs << "name" << "BayesianEstimator";

	fs << "method" << m_method;
	fs << "illuminants" << "[";
	for (unsigned int i = 0; i < m_illuminants.size(); i++) {
		fs << m_illuminants[i];
	}
	fs << "]";
	fs << "log_reflectance_prior" << m_logReflectancePrior;
	fs << "size" << m_size;

	return true;
}

bool BayesianEstimator::load(const std::string& filename)
{
	cv::FileStorage fs(filename, cv::FileStorage::READ);

	if (!fs.isOpened()) {
		return false;
	}

	std::string name = fs["name"];

	if (name != "BayesianEstimator") {
		return false;
	}

	m_method = (Method) (int) fs["method"];
	m_illuminants.clear();
	cv::FileNode illuminants = fs["illuminants"];
	if (illuminants.type() == cv::FileNode::SEQ) {
		for (unsigned int i = 0; i < illuminants.size(); i++) {
			CV_Assert(illuminants[i].type() == cv::FileNode::SEQ && illuminants[i].size() == 3);
			m_illuminants.push_back(cv::Vec3d((double) illuminants[i][0], (double) illuminants[i][1], (double) illuminants[i][2]));
		}
	}
	fs["log_reflectance_prior"] >> m_logReflectancePrior;
	m_size = fs["size"];

	return true;
}

int BayesianEstimator::error() {
	return 0;
}

} // namespace illuminantestimators
